from abc import ABC, abstractmethod
from typing import Type, Union

import pandas as pd

from format_call import OUTPUT_FUNCTION_MAPPING
from constants import BASE_URL, DCAT_URL


class BaseApi(ABC):
    '''
    Define the structure of APIs functionnality
    All APIs should follow this structure
    '''
    def __init__(self, export_format='JSON', parameters=None, root_url=BASE_URL) -> None:
        self.root_url = root_url
        self.export_format = export_format
        self.request_params = parameters
        
    @abstractmethod
    def set_url(self) -> None:
        pass
    
    def call_api(self) -> Union[dict, pd.DataFrame]:
        api_function = OUTPUT_FUNCTION_MAPPING[self.export_format]
        return api_function(f'{self.root_url}/{self.api_url}', self.request_params)
    

def call_api(class_type: Type[BaseApi], *args) -> Union[dict, pd.DataFrame]:
    '''
    Instantiate and call the class of the api
    '''
    api_class = class_type(*args)
    return api_class.call_api()


###############################################################
########################## i14y APIs ##########################
###############################################################

class CodeList(BaseApi):
    '''
    Export a codelist
    '''
    def __init__(
        self, 
        identifier: str, 
        export_format: str, 
        version_format: float, 
        annotations: bool
    ) -> None:
        super().__init__(export_format, f'annotations={str(annotations).lower()}')
        self.set_url(identifier, export_format, version_format)
        
    def set_url(self, identifier: str, export_format: str, version_format: float) -> None:
        self.api_url = f'api/CodeLists/{identifier}/exports/{export_format}/{version_format}'

        
class ContentConfiguration(BaseApi):
    '''
    Get the display information for the available configured content
    '''
    def __init__(self) -> None:
        super().__init__()
        self.set_url()
        
    def set_url(self) -> None:
        self.api_url = 'api/ContentConfigurations'

        
class ContentConfigurationIdentifier(BaseApi):
    '''
    Get a nomenclature information based on its identifier
    '''
    def __init__(self, identifier: str) -> None:
        super().__init__()
        self.set_url(identifier)
        
    def set_url(self, identifier: str) -> None:
        self.api_url = f'api/ContentConfigurations/{identifier}'

        
class NomenclaturePathNodes(BaseApi):
    '''
    Get the nodes of a path within a nomenclature
    '''
    def __init__(self, identifier: str, path: str, filters: dict, language: str) -> None:
        super().__init__('JSON', filters)
        self.set_url(identifier, path, language)
        
    def set_url(self, identifier: str, path: str, language: str) -> None:
        self.api_url = f'api/Nomenclatures/Childnodes/{identifier}/{language}/{path}'


class NomenclatureOneLevel(BaseApi):
    '''
    Get one level of a nomenclature
    '''
    def __init__(self, identifier: str, parameters: dict) -> None:
        super().__init__('CSV', parameters)
        self.set_url(identifier)
        
    def set_url(self, identifier: str) -> None:
        self.api_url = f'api/Nomenclatures/{identifier}/levelexport/CSV'


class NomenclatureMultipleLevels(BaseApi):
    '''
    Get multiple levels of a nomenclature
    '''
    def __init__(self, identifier: str, parameters: dict) -> None:
        super().__init__('CSV', parameters)
        self.set_url(identifier)
        
    def set_url(self, identifier: str) -> None:
        self.api_url = f'api/Nomenclatures/{identifier}/multiplelevels/CSV'


class NomenclatureSearch(BaseApi):
    '''
    Search a query within a nomenclature
    '''
    def __init__(self, identifier: str, parameters: dict) -> None:
        super().__init__('JSON', parameters)
        self.set_url(identifier)
        
    def set_url(self, identifier: str) -> None:
        self.api_url = f'api/Nomenclatures/{identifier}/search'

###############################################################
########################## dcat APIs ##########################
###############################################################

class AgentsList(BaseApi):
    '''
    List all agents
    '''
    def __init__(self) -> None:
        super().__init__('JSON', root_url=DCAT_URL)
        self.set_url()
        
    def set_url(self) -> None:
        self.api_url = 'api/Agent'

        
class AgentId(BaseApi):
    '''
    Return the agent with the corresponding agent id
    '''
    def __init__(self, agent_id: str) -> None:
        super().__init__('JSON', root_url=DCAT_URL)
        self.set_url(agent_id)
        
    def set_url(self, agent_id: str) -> None:
        self.api_url = f'api/Agent/{agent_id}'
        
        
class DatasetsList(BaseApi):
    '''
    List all datasets
    '''
    def __init__(self) -> None:
        super().__init__(root_url=DCAT_URL)
        self.set_url()
        
    def set_url(self) -> None:
        self.api_url = 'api/Dataset'
        
        
class DatasetIdDistributions(BaseApi):
    '''
    Return all distributions for the dataset with the corresponding dataset id
    '''
    def __init__(self, dataset_id: str) -> None:
        super().__init__(root_url=DCAT_URL)
        self.set_url(dataset_id)
        
    def set_url(self, dataset_id: str) -> None:
        self.api_url = f'api/Dataset/{dataset_id}/distributions'

        
class DatasetId(BaseApi):
    '''
    Return the dataset with the corresponding id
    '''
    def __init__(self, dataset_id: str) -> None:
        super().__init__(root_url=DCAT_URL)
        self.set_url(dataset_id)
        
    def set_url(self, dataset_id: str) -> None:
        self.api_url = f'api/Dataset/{dataset_id}'

        
class DatasetIdentifier(BaseApi):
    '''
    Return the dataset with the corresponding identifier
    '''
    def __init__(self, identifier: str) -> None:
        super().__init__(root_url=DCAT_URL)
        self.set_url(identifier)
        
    def set_url(self, identifier: str) -> None:
        self.api_url = f'api/Datataset/identifier/{identifier}'
        
        
class DatasetIdentifierDistributions(BaseApi):
    '''
    Return all distributions for the dataset with the corresponding dataset identifier
    '''
    def __init__(self, identifier: str) -> None:
        super().__init__(root_url=DCAT_URL)
        self.set_url(identifier)

    def set_url(self, identifier: str) -> None:
        self.api_url = f'api/Datataset/identifier/{identifier}/distributions'
        
        
class DistributionsList(BaseApi):
    '''
    List all distributions
    '''
    def __init__(self) -> None:
        super().__init__(root_url=DCAT_URL)
        self.set_url()
        
    def set_url(self) -> None:
        self.api_url = f'api/Distribution'

        
class DistributionId(BaseApi):
    '''
    Return the distribution with the corresponding id
    '''
    def __init__(self, distribution_id: str) -> None:
        super().__init__(root_url=DCAT_URL)
        self.set_url(distribution_id)
        
    def set_url(self, distribution_id: str) -> None:
        self.api_url = f'api/Distribution/{distribution_id}'
